# Copyright (c) 2016, Yahoo Inc.
# Copyrights licensed under the BSD License
# See the accompanying LICENSE.txt file for terms.

"""
Functions for handling the configuration settings
"""
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
import getpass
import io
import logging
import os
import tempfile
# noinspection PyUnresolvedReferences,PyPackageRequirements
from six.moves.configparser import ConfigParser
from jinja2 import Template

from .package import package_scripts_directory
from .plugin import config_defaults, config_types, config_update
from .utility import str_to_bool, str_to_list, str_format_env
from .virtualenv import default_virtualenv_directory


logger = logging.getLogger(__name__)  # pylint: disable=C0103


# Config file that is part of the package
PACKAGE_DEFAULT_CONFIG = os.path.join(
    os.path.dirname(__file__), 'deploy_default.conf'
)

CONFIG_PATH = [
    PACKAGE_DEFAULT_CONFIG,

    # Any deploy.conf files in the current directory
    'deploy.conf'
]


def cast_types(configuration, types_dict=None):
    """
    Update in place the configuration dictionary with inferred values if they
    aren't specified.

    Parameters
    ----------
    configuration : dict
    config_types : dict, optional
        Types dictionary, defaults to value from config_types()
    """
    if not types_dict:
        types_dict = config_types()

    for section, settings in configuration.items():
        for key, value in settings.items():
            try:
                setting_type = types_dict[section][key]
                if setting_type is bool:
                    setting_type = str_to_bool
                elif setting_type is list:
                    setting_type = str_to_list
                configuration[section][key] = setting_type(value)
                logger.debug(
                    'Forced value %r of type %r to type %r, result is %r',
                    value, type(value), setting_type,
                    configuration[section][key]
                )
            except (KeyError, TypeError):
                pass


def get_configuration(configuration=None):
    """
    Parse a configuration file

    Parameters
    ----------
    configuration : str or list, optional
        A configuration file or list of configuration files to parse,
        defaults to the deploy_default.conf file in the package and
        deploy.conf in the current working directory.

    Returns
    -------
    configparser
        The parsed configuration
    """
    if not configuration:  # pragma: no cover
        configuration = [
            # Config file that is part of the package
            # PACKAGE_DEFAULT_CONFIG,

            # Any deploy.conf files in the current directory
            'deploy.conf'
        ]
    config = ConfigParser()

    # Set the config defaults
    try:
        config.read_string(config_defaults())
    except AttributeError:
        config.readfp(io.BytesIO(config_defaults()))

    logger.debug('Working with default dict: %r', config_defaults())
    config.read(configuration)
    return config


def get_configuration_dict(configuration=None, value_types=None):
    """
    Parse the configuration files

    Parameters
    ----------
    configuration : str or list, optional
        A configuration file or list of configuration files to parse,
        defaults to the deploy_default.conf file in the package and
        deploy.conf in the current working directory.

    value_types : dict, optional
        Dictionary containing classes to apply to specific items

    Returns
    -------
    dict
        Configuration dictionary
    """
    if not value_types:  # pragma: no cover
        value_types = config_types()

    config = get_configuration(configuration)

    result_dict = {}
    for section in config.sections():
        result_dict[section] = {}
        for key, val in config.items(section):
            result_dict[section][key] = str_format_env(val)

    config_update(result_dict)

    if 'locations' not in result_dict.keys():
        result_dict['locations'] = {}
    result_dict['locations']['package_scripts'] = package_scripts_directory()
    if not result_dict['global'].get('virtualenv_dir', None):
        result_dict['global']['virtualenv_dir'] = \
            default_virtualenv_directory()

    cast_types(result_dict)

    return result_dict


def generate_parsed_config_file(source=None, dest=None):
    """
    Generate a conf file with the environment variables expanded

    Parameters
    ----------
    source : str, optional
        The path to the source file

    dest : str, optional
        The path to the desired destination file

    Returns
    -------
    str
        Path to the generated config file
    """
    config_data = None
    if not source:  # pragma: no cover
        source = 'deploy.conf'

    if isinstance(source, list):
        source = source[0]

    with open(source) as config_data_handle:
        config_data = config_data_handle.read()

    if not config_data:  # pragma: no cover
        return None

    template = Template(config_data)
    result = template.render(**os.environ)

    if not dest:
        with tempfile.NamedTemporaryFile(
            delete=False, suffix='.conf'
        ) as dest_handle:
            dest_handle.write(result.encode())
            dest = dest_handle.name
    else:
        with open(dest, 'w') as new_conf_handle:
            new_conf_handle.write(result)

    return dest


def parse_arguments(configuration=None):
    """
    Parse the command line arguments

    Parameters
    ----------
    configuration : list, optional
        A list of config files to parse, defaults to the packaged
        deploy_default.conf and the deploy.conf file in the current
        directory.

    Returns
    -------
    argparse namespace object
        With the configuration based on the passed command line arguments
        and defaults from the deploy_default.conf file in the package.
    """
    config = get_configuration_dict(
        configuration,
        config_types()
    )
    virtualenvuser = config['global']['virtualenv_user']
    if not virtualenvuser:  # pragma: no cover
        virtualenvuser = getpass.getuser()

    parser = ArgumentParser(
        description='Deploy a python application into a virtualenv',
        formatter_class=ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        'name', nargs='?', default=config['global']['name'],
        help='VirtualEnv name'
    )
    parser.add_argument(
        '--python',
        '-p',
        default=config['global']['basepython'],
        help='The Python interpreter to use, e.g., --python=python3.5 '
             'will use the python3.5 interpreter to create the new '
             'environment.  The default is the interpreter that virtualenv '
             'was installed with.'
    )
    parser.add_argument(
        '--requirement',
        '-r',
        default=[],
        action='append',
        help='Install from the given requirements file. This option can be '
             'used multiple times.'
    )
    parser.add_argument(
        '--virtualenvdir',
        default=config['global']['virtualenv_dir'],
        help='Directory to build the virtualenv'
    )
    parser.add_argument(
        '--virtualenvuser',
        default=virtualenvuser,
        help='The user to create the virtualenv'
    )
    parser.add_argument(
        '--virtualenvgroup',
        default=config['global']['virtualenv_group'],
        help='The group to create the virtualenv'
    )
    parser.add_argument(
        '--virtualenvversion_package',
        default=config['global']['virtualenv_version_package'],
        help='Version the virtualenv based on the version of a package'
    )
    parser.add_argument(
        '--install_os_packages',
        default=config['global']['install_os_packages'],
        help='Install OS packages'
    )
    parser.add_argument(
        '--verbose', action='store_true', default=False,
        help='More verbose output'
    )
    parser.add_argument(
        '--upgrade', action='store_true', default=False,
        help='Upgrade packages when installing'
    )
    return parser.parse_args()



# Copyright (c) 2016, Yahoo Inc.
# Copyrights licensed under the BSD License
# See the accompanying LICENSE.txt file for terms.

"""
Functions to enable packaging plugin functionality
"""
import logging
import os
import shutil
import subprocess
import sys
from jinja2 import Template
from .config import get_configuration_dict, get_configuration, generate_parsed_config_file
from .contextmanager import InTemporaryDirectory, working_dir
from .utility import find_executable, update_recursive, csv_list


logger = logging.getLogger(__name__)  # pylint: disable=C0103


class InvirtualenvPlugin(object):
    package_formats = []
    config_default = ""
    config_types = {}
    default_config_filename = 'invirtualenv.configuration'
    package_template = ''
    hash = None  # PIP hash algorithm to use, can be sha256, sha384, sha512 or None (no hashing)

    def __init__(self, config_file='deploy.conf'):
        self.config_file = config_file
        self.config = get_configuration_dict(configuration=config_file)
        self.loaded_configuration = get_configuration(configuration=config_file)

    # Methods that need to be written for each plugin type
    def run_package_command(self, package_hashes, wheel_dir='wheels'):
        """
        Run the command to generate the package based on the hash
        """
        pass

    def system_requirements_ok(self):
        """
        Check if all the system requirements for this plugin are met.

        Returns
        -------
        bool
            True if requirements are met, False otherwise
        """
        return True

    @property
    def pip_cmd(self):
        """
        Get the pip command used to create wheels of packages.

        The intention is to use the same version of pip to build the wheels
        as would be used to deploy them.

        The full path to the python interpreter is used to avoid shebang
        line length issues.

        Returns
        -------
        list
            Command that can be used to invoke pip
        """
        basepython = self.config['global'].get('basepython', 'python3')
        python_executable = find_executable(basepython)
        if not python_executable:
            python_executable = sys.executable
        bin_dir = os.path.dirname(python_executable)
        return [python_executable, '-m', 'pip']

    def supported_formats(self):
        """
        Formats supported by this plugin that can be generated on this system

        Returns
        -------
        list
            Package formats that this plugin supports
        """
        if self.system_requirements_ok():
            return self.package_formats
        return []

    def create_package(self, package_type):
        """
        Generate a package of the specified type

        Parameters
        ----------
        str: package_type
            The type of package to generate
        """
        if package_type not in self.supported_formats():
            return None

        original_directory = os.getcwd()
        with InTemporaryDirectory():
            tempdir = os.getcwd()
            wheel_dir = 'wheels'
            os.makedirs(wheel_dir)
            hashes = self.generate_wheel_packages(wheel_dir)
            self.generate_wheel_archive()
            deps = []
            for package_name, package_hash in hashes.items():
                deps.append('{package_name} --hash={package_hash}'.format(package_name=package_name, package_hash=package_hash))
            self.config['pip']['deps'] = deps
            if self.hash:
                self.loaded_configuration['pip']['deps'] = '\n'.join(deps)
            with open('deploy.conf.unparsed', 'w') as deploy_conf_handle:
                self.loaded_configuration.write(deploy_conf_handle)
            generate_parsed_config_file('deploy.conf.unparsed', 'deploy.conf')
            package = self.run_package_command(hashes, wheel_dir=wheel_dir)  # pylint: disable=E1128,E1111
            if package and os.path.exists(package):
                source = package
                dest = os.path.join(original_directory, os.path.basename(package))
                shutil.copyfile(source, dest)
                return dest
            return package

    def generate_wheel_archive(self, filename=None):
        if not filename:
            filename = 'wheels.tar.gz'
        subprocess.check_call(['tar', '-czf', filename, 'wheels'])

    def generate_wheel_packages(self, wheeldir):
        """
        Generate wheel packages for all dependencies

        Parameters
        ----------
        wheeldir: str
            The directory path to store the generated wheel packages

        Returns
        -------
        dict of filename, pip requirements line
        """
        if not self.config['pip'].get('deps'):
            return {}
        hashes = {}
        with working_dir(wheeldir):
            deps = self.config['pip'].get('deps', []) + ['invirtualenv']
            cmd = self.pip_cmd + ['-q', 'wheel', '-w', '.'] + deps
            logger.debug('Running pip command %r to generate wheel packages', cmd)
            subprocess.check_call(cmd)
            for filename in os.listdir('.'):
                if filename.endswith('.whl'):
                    cmd = self.pip_cmd + ['hash']
                    if self.hash:
                        cmd += ['-a', self.hash]
                    cmd += [filename]
                    logger.debug('Running pip command %r to generate package hash for %r', cmd, filename)
                    hashes[filename] = '='.join(subprocess.check_output(cmd).decode().split(os.linesep)[1].split('=')[1:])
                    logger.debug('Got requirements line %r', hashes[filename])
        return hashes

    def render_template_with_config(self, template_str=None):
        if not template_str:
            template_str = self.package_template
            template = Template(template_str)
            return template.render(self.config)



# Copyright (c) 2016, Yahoo Inc.
# Copyrights licensed under the BSD License
# See the accompanying LICENSE.txt file for terms.

"""
Functions to enable packaging plugin functionality
"""
import logging
import os
import pkg_resources
from .utility import find_executable, update_recursive, csv_list


logger = logging.getLogger(__name__)  # pylint: disable=C0103


CONFIG_DEFAULT = """[global]
name =
basepython =
description = No description is available
install_manifest =
install_os_packages = False
version =
virtualenv_dir =
virtualenv_deploy_dir=
virtualenv_version_package =
virtualenv_user =
virtualenv_group =

[pip]
pip_version =
deps:

[rpm]
fail_missing_yum = True
deps:

"""
CONFIG_TYPES = {
    'global': {
        'install_manifest': csv_list,
        'install_os_packages': bool,
    },
    'pip': {
        'deps': list
    },
    'rpm': {
        'deps': list,
        'fail_missing_yum': bool
    },
}


def installed_plugins():
    plugins = []
    for entry_point in pkg_resources.iter_entry_points(group='invirtualenv.plugin'):
        plugins.append(entry_point.load())
    return plugins


def package_formats():
    """
    Get a list of all the supported package types

    Returns
    -------
    list
        A list of supported package type strings
    """
    supported_types = []
    for plugin in installed_plugins():
        supported_types += plugin().supported_formats()

    # Legacy plugin support
    for entry_point in pkg_resources.iter_entry_points(
            group='invirtualenv.supported'
    ):  # pragma: no cover
        supported = entry_point.load()
        logger.debug(supported)
        supported_types += supported()

    supported_types.sort()
    logger.debug('Supported package types: %s', supported_types)
    return supported_types


def config():
    """
    Get default configuration settings for all installed plugins

    Returns
    -------
    str
        Configparser format configuration as a string
    dict
        Configuration types dictionary
    """
    config_string = CONFIG_DEFAULT
    config_types_dict = CONFIG_TYPES

    for plugin in installed_plugins():
        if plugin.config_default:
            config_string += plugin.config_default + os.linesep
        if plugin.config_types:
            config_types_dict = update_recursive(
                config_types_dict, plugin.config_types
            )

    # Legacy plugin support
    for entry_point in pkg_resources.iter_entry_points(
            group='invirtualenv.config'
    ):  # pragma: no cover
        default_config_function = entry_point.load()
        default_config, default_types = default_config_function()
        if default_config:
            logger.debug('Adding to default config: %r', default_config)
            config_string += default_config
        if default_types:
            logger.debug('Adding to default types: %r', default_types)
            config_types_dict = update_recursive(
                config_types_dict, default_types
            )

    return config_string, config_types_dict


def config_update(configuration):
    """
    Allow plugins to update the configuration after the values have been
    propagated.

    This function updates the dictionary in place.

    Parameters
    ----------
    configuration : dict
        The configuration dictionary
    """
    for entry_point in pkg_resources.iter_entry_points(
            group='invirtualenv.config_update'
    ):  # pragma: no cover
        config_update_function = entry_point.load()
        config_update_function(configuration)

    logger.debug('Updated configuration: %r', configuration)


def config_defaults():
    """
    Get default configuration settings for all installed plugins

    Returns
    -------
    str
        ConfigParser format configuration as a string
    """
    return config()[0]


def config_types():
    """
    Get the default types for all configuration valuse based on installed
    plugins.

    Returns
    -------
    dict
        Dictionary with configuration values as keys and object classes as
        values
    """
    return config()[1]


def get_package_plugin(package_type):
    """
    Get a plugin for a specific package

    Parameters
    ----------
    package_type: str
        The package type to fetch

    Returns
    -------
    InvirtualEnvPlugin:
        The invirtualenv plugin for the specific package_type
    """
    for plugin in installed_plugins():
        if package_type in plugin.package_formats:
            return plugin


def create_package_configuration(package_type):
    """
    Create a package of a specific package type

    This functions iterates all the registered invirtualenv.create_package
    entry points and runs them passing the package_type argument to them.

    These plugins should return None if they do not handle that package type.

    Parameters
    ----------
    package_type : str
        The package type to create a package for

    """
    for plugin in installed_plugins():
        if package_type in plugin.package_formats:
            return plugin().render_template_with_config()


def create_package(package_type):
    """
    Create a package of a specific package type

    This functions iterates all the registered invirtualenv.create_package
    entry points and runs them passing the package_type argument to them.

    These plugins should return None if they do not handle that package type.

    Parameters
    ----------
    package_type : str
        The package type to create a package for

    """
    for plugin in installed_plugins():
        package_name = plugin().create_package(package_type)
        if package_name:
            return package_name

    # Legacy plugin support
    for entry_point in pkg_resources.iter_entry_points(
            group='invirtualenv.create_package'
    ):
        package = entry_point.load()
        package_name = package(package_type)
        if package_name:
            return package(package_type)



# Copyright (c) 2016, Yahoo Inc.
# Copyrights licensed under the BSD License
# See the accompanying LICENSE.txt file for terms.

"""
Functions for creating and managing python virtual environments
"""
from __future__ import print_function
import getpass
import hashlib
import json
import logging
import os
from pwd import getpwnam
import shutil
import subprocess
import sys

try:
    import venv
    BUILTIN_VENV = True
except ImportError:
    BUILTIN_VENV = False

from .exceptions import BuildException
from .utility import chown_recursive, which


logger = logging.getLogger(__name__)  # pylint: disable=C0103


def default_virtualenv_directory():
    """
    Get the default virtualenv directory for the current system/platform

    Returns
    -------
    str
        The path to the default virtualenv directory
    """
    if os.path.exists('/var/tmp'):
        return '/var/tmp/virtualenv'

    return '/tmp/virtualenv'


def virtualenv_command(install_virtualenv=False):
    """
    Return the virtualenv command

    Parameters
    ----------
    install_virtualenv : bool
        If True, Install virtualenv if necessary.  default=False

    Returns
    -------
    str
        Path to the virtualenv command
    """
    if not hasattr(sys, 'frozen'):
        if install_virtualenv:
            packages = os.popen('pip freeze').read()
            if 'virtualenv' not in packages:
                subprocess.check_output(['pip', 'install', 'virtualenv'])
    return which('virtualenv')


def virtualenv_bin_file_hashes(virtualenv_dir):
    """
    Calculate a hash of all files in the virtualenv bin directory
    Parameters
    ----------
    virtualenv_dir : str
        The root directory of the virtualenv to operate on
    Returns
    -------
    dict:
        Key = Filename, value = hash
    """
    bin_dir = os.path.join(virtualenv_dir, 'bin')
    if not os.path.exists(bin_dir):
        return
    result = {}
    for filename in os.listdir(bin_dir):
        full_filename = os.path.join(bin_dir, filename)
        if os.path.isdir(full_filename):
            continue
        filehash = hashlib.sha256()
        with open(full_filename, 'rb') as handle:
            filehash.update(handle.read())
        result[filename] = filehash.hexdigest()
    return result


def remove_virtualenv(name, directory=None):
    """
    Remove a virtualenv from a directory
    :param name:
    :param directory:
    :return:
    """
    if not directory:
        directory = '.'

    venv_directory = os.path.expanduser(os.path.join(directory, name))
    if os.path.exists(venv_directory):
        logger.debug('Removing virtualenv directory %r', venv_directory)
        shutil.rmtree(venv_directory)


def upgrade_package_tools(virtualenv_directory, verbose=False):
    """
    Upgrade the packages used to install/build packages in the virtualenv
    Parameters
    ----------
    virtualenv_directory: str
        The directory that contains the virtualenv
    """
    python_interpreter = os.path.join(virtualenv_directory, 'bin/python')
    pip_command = os.path.join(virtualenv_directory, 'bin/pip')

    for command in [
        [python_interpreter, pip_command, 'install', '--upgrade', 'pip'],
        [python_interpreter, pip_command, 'install', '--upgrade', 'setuptools'],
        [python_interpreter, pip_command, 'install', '--upgrade', 'wheel'],
    ]:
        try:
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)
            if verbose:
                print(output.decode().strip())
        except subprocess.CalledProcessError as error:
            if verbose:
                print(error.output.decode().strip())
            logger.debug(error.output.decode().strip())
            error_message = 'Upgrade command {command} in virtualenv {virtualenv_directory} failed'.format(
                command=command,
                virtualenv_directory=virtualenv_directory
            )
            raise BuildException(error_message)


def build_virtualenv(
        name, directory, python_interpreter=None, user=None, verbose=False):
    """
    Build a virtualenv in a directory

    Parameters
    ----------
    name : str
        Name of the virtualenv to create

    directory : str
        Directory to create the virtualenv in

    python_interpreter : str, optional
        Python interpreter to provide in the virtualenv, defaults to the
        interpreter that is running the virtualenv command

    verbose : bool
        If True, provides status output while running.

    Returns
    -------
    str
        Full path to the root of the virtualenv directory

    Raises
    ------
    BuildException
        The Virtualenv build failed
    """

    # if not python_interpreter:
    #     if not hasattr(sys, 'frozen'):
    #         python_interpreter = sys.executable
    # logger.debug('Python interpreter is: %s' % sys.executable)
    cwd = os.getcwd()
    if not os.path.isdir(directory):
        os.makedirs(directory)

    virtualenv_dir = os.path.join(directory, name)

    if False and not python_interpreter and BUILTIN_VENV and \
            not hasattr(sys, 'frozen'):
        logger.debug(
            'Building virtualenv %r using the built in venv module',
            virtualenv_dir
        )
        venv.create(virtualenv_dir, with_pip=True)
    else:
        os.chdir(directory)
        command = [virtualenv_command()]
        if python_interpreter:
            command += ['-p', python_interpreter]
        command += [name]
        logger.debug(
            'Building virtualenv using external command %r', ' '.join(command)
        )
        try:
            output = subprocess.check_output(
                command,
                stderr=subprocess.STDOUT,
            )
            if verbose:
                print(output.decode().strip())
        except subprocess.CalledProcessError as error:
            if verbose:
                print(error.output.decode().strip())
            logger.debug(error.output.decode().strip())
            logger.exception(
                'Virtualenv create command %r failed', ' '.join(command))
            remove_virtualenv(name, directory)
            os.chdir(cwd)
            raise BuildException('Virtualenv create failed')
        os.chdir(cwd)

    for directory_name in ['conf', 'logs']:
        filename = os.path.join(virtualenv_dir, directory_name)
        if not os.path.exists(filename):
            logger.debug('Creating %r directory', filename)
            os.makedirs(filename)

    upgrade_package_tools(virtualenv_dir, verbose=verbose)

    before_binfiles_filename = os.path.join(virtualenv_dir, 'conf/binfiles_predeploy.json')
    with open(before_binfiles_filename, 'w') as before_binfiles_handle:
        before_binfiles = virtualenv_bin_file_hashes(virtualenv_dir)
        before_binfiles_handle.write(json.dumps(before_binfiles))

    return virtualenv_dir


def install_requirements(
        requirements, virtualenv, user=None, upgrade=False, verbose=False,
        pip_version=None, use_index=True, use_local_wheels=False
):
    """
    Open one or more requirements files and run pip -r to install them

    Parameters
    ----------
    requirements : str
        Filename containing requirements

    virtualenv : str
        Full path to the virtualenv to install into

    user : str, optional
        The user:group to run the install as

    upgrade : bool
        If True, tell pip to upgrade when running the install.  Default=False

    verbose : bool
        If True, provides status output while running.

    pip_version: str, optional
        Install the requirements with the specified version of pip

    use_index : bool, optional
        Allow pip to use an external index
        Default=True

    use_local_wheels: bool, optional
        Install wheels from local directory
        Default=False
    """
    logger.info(
        'Installing requirements from requirements file: %r '
        'into virtualenv %r as user %r',
        requirements, virtualenv, user
    )
    logger.debug('Current user is: %r', getpass.getuser())
    logger.debug(
        'Current uid: %d, Effective uid: %d', os.getuid(), os.geteuid()
    )

    if isinstance(requirements, str):
        requirements = [requirements]

    extra_pip_args = []
    root_logger = logging.getLogger()
    if upgrade:
        extra_pip_args.append('-U')
    if not verbose and root_logger.level > logging.DEBUG:
        extra_pip_args.append('-q')
    if not use_index:
        extra_pip_args.append('--no-index')
    if use_local_wheels:
        # Get the wheels_dir path from virtualenv path. Instead of downloading
        # packages from pypi we will be installing wheels from local dir.
        package_data_dir_name = os.path.basename(virtualenv)
        package_data_dir = os.path.join("/usr/share/", package_data_dir_name)
        wheels_dir = os.path.join(package_data_dir, "wheels")
        extra_pip_args += ['--find-links', wheels_dir]

    user_uid = None
    user_gid = None
    if user:
        user_uid = getpwnam(user).pw_uid
        user_gid = getpwnam(user).pw_gid
        logger.debug('Installing pip requirements as %r', user)

    virtualenv_bin = os.path.join(virtualenv, 'bin')
    pip_cache_dir = os.path.join(virtualenv, '.cache')
    pip_cache_dir = '~/.cache'
    if user_uid:
        chown_recursive('pip_cache_dir', user_uid, user_gid)

    for requirement in requirements:
        logger.info('Installing python requirements from file %r', requirement)
        command = [
            os.path.join(virtualenv_bin, 'pip'),
            'install',
            '--cache-dir', pip_cache_dir,
            '-r', requirement,
        ] + extra_pip_args
        logger.debug('Running command: %s', ' '.join(command))
        try:
            output = subprocess.check_output(
                command, stderr=subprocess.STDOUT,
                # preexec_fn=change_uid_gid(user_uid=user_uid)
            )
            if verbose:
                print(output.decode())
        except subprocess.CalledProcessError as error:
            print(error.output.decode())
            logger.exception('PIP install operation failed')
            raise BuildException('PIP install operation failed')

    after_binfiles_filename = os.path.join(
        virtualenv, 'conf/binfiles_postdeploy.json'
    )
    logger.debug('Writing binfiles hashes to %r', after_binfiles_filename)
    with open(after_binfiles_filename, 'w') as after_binfiles_handle:
        after_binfiles = virtualenv_bin_file_hashes(virtualenv)
        after_binfiles_handle.write(json.dumps(after_binfiles))



"""
Command line interface to invirtualenv v2
"""
import argparse
import logging
import os
import shutil
import sys
from .config import get_configuration_dict
from .contextmanager import InTemporaryDirectory
from .exceptions import PackageGenerationFailure
from .plugin import create_package, create_package_configuration, get_package_plugin, package_formats


LOGGER = logging.getLogger(__name__)


def parse_cli_arguments():

    # Determine the package types we can create on the current system
    package_choices = package_formats()

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--deploy_conf', default='deploy.conf', help='Deploy configuration filename or url')
    command_parser = parser.add_subparsers(title='command', dest='command')
    list_plugins_parser = command_parser.add_parser('list_plugins', help='List the installed invirtualenv plugins')
    package_config_parser = command_parser.add_parser('create_package_config', help='Generate the packaging configuration file')
    package_config_parser.add_argument('package_type', choices=package_choices, help='Type of package to create')
    package_config_parser.add_argument('--outfile', '-o', default=None, help='Output file name')

    package_create_parser = command_parser.add_parser('create_package', help='Generate a package from a deployment configuration')
    package_create_parser.add_argument('package_type', choices=package_choices, help='Type of package to create')

    get_setting_parser = command_parser.add_parser('get_setting', help='Get a setting value from the configuration')
    get_setting_parser.add_argument('section', help="the configuration section to get the setting from")
    get_setting_parser.add_argument('item', help='The item to get from the configuration')
    args = parser.parse_args()
    return args


def get_setting_command(args):
    """
    Get the value of a setting in the deploy.conf

    Parameters
    ----------
    args: argparse.Namespace
        The argparse parser namespace with the parsed cli settings

    Returns
    -------
    str:
        The value of the setting or None
    """
    LOGGER.debug('Getting value for %s=%s', args.section, args.item)
    rc = 0
    config = get_configuration_dict([args.deploy_conf])
    try:
        value = config[args.section][args.item]
    except KeyError:
        value = ''
        rc = 1
    LOGGER.debug('Got value %r for %s=%s', value, args.section, args.item)
    return rc, value


def create_config_command(args):
    if not package_formats():  # pragma: no cover
        raise PackageGenerationFailure('No supported package creation plugins found')

    outfile = args.outfile
    if not outfile:
        plugin = get_package_plugin(args.package_type)
        outfile = plugin.default_config_filename

    deploy_config_contents = ''
    with open(args.deploy_conf) as deploy_conf_handle:
        deploy_config_contents = deploy_conf_handle.read()

    with InTemporaryDirectory():
        with open('deploy.conf', 'w') as deploy_conf_handle:
            deploy_conf_handle.write(deploy_config_contents)

    result = create_package_configuration(args.package_type)
    if outfile:
        with open(outfile, 'w') as output_handle:
            output_handle.write(result)
    else:
        sys.stdout.write(result)

    return 0, outfile


def create_package_command(args):
    if not package_formats():
        raise PackageGenerationFailure('No supported package creation plugins found')

    deploy_config_contents = ''
    with open(args.deploy_conf) as deploy_conf_handle:
        deploy_config_contents = deploy_conf_handle.read()

    orig_directory = os.getcwd()
    with InTemporaryDirectory():
        with open(args.deploy_conf, 'w') as deploy_conf_handle:
            deploy_conf_handle.write(deploy_config_contents)
        package_file = create_package(args.package_type)
        if package_file:
            dest_package_file = os.path.join(orig_directory, os.path.basename(package_file))
            if os.path.exists(package_file):
                shutil.copy(package_file, dest_package_file)

    if package_file:
        logging.debug('Generated package file: %s' % dest_package_file)
        return 0, 'Generated package file:' + dest_package_file

    raise PackageGenerationFailure('Unable to generate a package file using the %r plugin' % args.package_type)


def list_plugins_command(args):
    installed_plugins = package_formats()

    if not installed_plugins:
        return 0, 'No plugins are installed'
    result = 'Installed plugins:' + os.linesep
    for plugin in package_formats():
        result += '\t' + plugin + os.linesep
    return 0, result


def main(test=False):
    logging.basicConfig(level=logging.INFO)
    args = parse_cli_arguments()

    rc = 0
    output = ''
    if args.command in ['create_package']:
        rc, output = create_package_command(args)
    elif args.command in ['create_package_config']:
        rc, output = create_config_command(args)
    elif args.command in ['list_plugins']:
        rc, output = list_plugins_command(args)
    elif args.command in ['get_setting']:
        rc, output = get_setting_command(args)
    if test:
        return rc, output

    print(output)
    sys.exit(rc)



#!/usr/bin/env python3
from __future__ import print_function
import logging
import os
import shutil
import sys

from configparser import ConfigParser, NoOptionError
from invirtualenv.deploy import build_deploy_virtualenv


def update_config(venv_dir, config_filename=None):
    if not config_filename:
        config_filename = 'deploy.conf'
    config = ConfigParser()
    config.read(config_filename)
    config.set('global', 'virtualenv_deploy_dir', venv_dir)
    with open(config_filename, 'w') as configfile:
        config.write(configfile)


def get_config_flag(section, option, config_filename=None):
    if not config_filename:
        config_filename = 'deploy.conf'
    config = ConfigParser()
    config.read(config_filename)
    try:
        return config.getboolean(section, option)
    except NoOptionError:
        return


if __name__ == "__main__":
    # logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger(sys.argv[0])

    logger.debug('Running post install steps, from directory %r' % os.getcwd())

    if not os.path.exists('deploy.conf'):
        print("No 'deploy.conf' found.  Doing nothing", file=sys.stderr)
        sys.exit(1)

    logger.debug('Using the following deploy.conf: %s' % open('deploy.conf').read())

    venv_directory = None
    try:
        venv_directory = build_deploy_virtualenv(update_existing=True)
        update_config(venv_directory)
    except Exception:
        print('Unable to create the python virtualenv', file=sys.stderr)
        logger.exception('The virtualenv create failed, removing venv_directory')
        if venv_directory and os.path.exists(venv_directory):
            print('Removing the %r virtualenv directory' % venv_directory)
            shutil.rmtree(venv_directory)
        sys.exit(1)

    try:
        if get_config_flag('global', 'link_bin_files'):
            logger.debug('Linking files in the virtualenv bin directory to %r', os.path.dirname(sys.executable))
            try:
                from invirtualenv.deploy import link_deployed_bin_files
                link_deployed_bin_files(venv_directory, '/usr/bin')
            except ImportError:
                print('WARNING: The installed version of invirtualenv does not support linking bin files')
    except Exception:
        logger.exception('An error occurred linking bin files into %r', os.path.dirname(sys.executable))

    print('Created virtualenv %r' % venv_directory)
    sys.exit(0)



# Copyright (c) 2016, Yahoo Inc.
# Copyrights licensed under the BSD License
# See the accompanying LICENSE.txt file for terms.

"""
Module to create/deploy a virtualenv
"""
from __future__ import print_function
import getpass
import json
import logging
import os
from pwd import getpwnam
import subprocess
import tempfile
from .utility import display_header
from .config import get_configuration_dict, parse_arguments
from .exceptions import AlreadyExists, BuildException, \
    InsufficientPermissions, NoPackageVersions
from .package import install_prereq_packages, latest_package_version
from .utility import which
from .virtualenv import build_virtualenv, install_requirements, \
    remove_virtualenv


logger = logging.getLogger(__name__)  # pylint: disable=C0103


COMMANDS = {
    'apt-get': '/usr/bin/apt-get',
    'pip': 'pip',
    'yum': '/usr/bin/yum'
}


def fix_file_ownership(virtualenv, user, group):
    """
    Fix the file ownership of a virtualenv

    Parameters
    ----------
    virtualenv : str
        Virtualenv path

    user : str
        User to change to

    group : str
        Group to change to
    :return:
    """
    if (user and group) and virtualenv:
        subprocess.check_call(
            [
                'chown',
                '-R',
                '%s:%s' % (
                    user,
                    group
                ),
                virtualenv
            ]
        )
        subprocess.check_call(
            ['chmod', '-R', 'g+wx', virtualenv]
        )


def install_python_dependencies(virtualenv, deps=None, requirements=None,
                                upgrade=False, verbose=False, pip_version=None,
                                use_index=True, use_local_wheels=False):
    """
    Install python dependencies from a requirements file or
    deploy.conf manifest

    Parameters
    ----------
    virtualenv : str
        The virtualenv to install into

    deps : list, optional
        A list of python packages to install

    requirements : str, optional
        The requirements.txt file to install from

    upgrade : bool, optional
        Tell pip to upgrade packages when installing, Default=False

    verbose : bool, optional
        Display command output when running the dependency operations,
        Default=False

    use_index : bool, optional
        Allow pip to use an external index
        Default=True

    use_local_wheels: bool, optional
        Install from local wheels directory instead of pypi
        Default=False

    Raises
    ------
    BuildException - If package installation fails
    """
    if not deps and not requirements:  # pragma: no cover
        if verbose:
            print('No requirements specified to install')
        return

    if deps:
        logger.debug('Installing dependencies from configuration deps: %s', ' '.join(deps))
        with tempfile.NamedTemporaryFile() as requirements_handle:
            requirements_handle.write('\n'.join(deps).encode())
            requirements_handle.flush()
            requirements_handle.seek(0)
            install_requirements(
                requirements_handle.name,
                virtualenv=virtualenv,
                upgrade=upgrade,
                verbose=verbose,
                pip_version=pip_version,
                use_index=use_index,
                use_local_wheels=use_local_wheels
            )
    if requirements:
        logger.debug('Installing dependencies from requirements file %r', requirements)
        install_requirements(
            requirements, virtualenv=virtualenv, upgrade=upgrade,
            verbose=verbose, pip_version=pip_version, use_index=use_index,
            use_local_wheels=use_local_wheels
        )


def install_rpm_dependencies(deps=None, fail_missing=True):  # pragma: no cover
    """
    Install rpm dependencies from deploy.conf manifest

    Parameters
    ----------
    deps : list, optional
        A list of rpm packages to install

    fail_missing : bool, optional
        Generate an exception and fail if the "yum" command is not available.
        default = True

    Raises
    ------
    BuildException - If package installation fails
    """
    if deps and not os.path.exists(COMMANDS['yum']):
        if fail_missing:
            raise BuildException('No %r command available' % COMMANDS['yum'])
        logger.warning('The %r command is missing', COMMANDS['yum'])

    if deps:
        logger.debug('Installing yum packages: %r', deps)
        if os.path.exists(COMMANDS['yum']):
            command = [COMMANDS['yum'], 'install', '-y'] + deps
            logger.debug('Running command: %s', ' '.join(command))
            subprocess.check_call(command)


def build_deploy_virtualenv(arguments=None, configuration=None, update_existing=True, verbose=None):
    """
    Build and deploy a python virtualenv

    Parameters
    ----------
    arguments : argparse namespace, optional
        An argparse namespace with all of the required command line arguments
        defaults to parsing the command line arguments if not provided.

    configuration : str or list, optional
        A configuration file or list of configuration files to parse,
        defaults to the deploy_default.conf file in the package and
        deploy.conf in the current working directory.

    update_existing : bool
        If True, allow updating an existing virtualenv, Otherwise generate
        an exception.  Default=True

    verbose : bool
        If True, provides status output while running.

    Raises
    ------
    AlreadyExists
        The virtualenv already exists

    BuildError
        The specified package manifest cannot be deployed to the virtualenv

    InsufficientPermissions
        The current user lacks permissions to complete the specified operation

    NoPackageVersions
        A virtualenv_version_package was specified in the configuration but
        no versions for that package where found on artifactory.
    """
    # display_header('Parsing the configuration')
    config = get_configuration_dict(configuration=configuration)
    if not arguments:
        logger.debug(
            'No arguments dictionary passed, parsing command line arguments')
        arguments = parse_arguments(configuration=configuration)
        logger.debug('Arguments are: %s', arguments)

    if verbose is None:
        verbose = arguments.verbose

    # If an interperter is specified, validate it exists
    if config['global']['basepython']:
        if not os.path.exists(config['global']['basepython']):
            config['global']['basepython'] = which(
                config['global']['basepython'])

    version = config['global']['version']
    if arguments.virtualenvversion_package:
        if verbose:
            display_header(
                'Getting version based on package %r from artifactory' %
                arguments.virtualenvversion_package
            )
        version = latest_package_version(arguments.virtualenvversion_package)
        if not version:
            raise NoPackageVersions(
                'Unable to find any package versions on for package '
                '%r' % arguments.virtualenvversion_package
            )

    if version:
        arguments.name += '_' + version

    logger.debug('Using version: ' + version)

    # In no package manifest installs where specified default to all
    if not config['global']['install_manifest']:
        config['global']['install_manifest'] = ['pip', 'rpm']

    # Make sure we have needed privileges
    if os.geteuid() != 0:
        if 'rpm' in config['global']['install_manifest']:
            if config['rpm']['deps']:
                raise InsufficientPermissions(
                    'Must run as root to install rpm packages'
                )
        if arguments.install_os_packages:
            raise InsufficientPermissions(
                'Must run as root to install python pre-req rpm packages'
            )

    # Make sure we aren't overwriting a virtualenv without specifying
    # update.
    virtualenv = os.path.join(arguments.virtualenvdir, arguments.name)
    if os.path.exists(virtualenv):
        if not update_existing:
            raise AlreadyExists(
                'Virtualenv %r already exists' % virtualenv
            )

    # Install the python OS rpm packages if requested.
    if arguments.install_os_packages:
        if verbose:
            display_header('Installing Operating System Pre-Req packages')
        install_prereq_packages()

    # Install rpm packages if specified
    if 'rpm' in config['global']['install_manifest']:
        if config['rpm']['deps']:
            if verbose:
                display_header('Installing rpm dependencies')
            install_rpm_dependencies(
                deps=config['rpm']['deps'],
                fail_missing=config['rpm']['fail_missing_yum']
            )

    # By default don't use local wheels.
    use_local_wheels = config['global'].get('use_local_wheels', False)

    if verbose:
        display_header('Building virtualenv')
    virtualenv = build_virtualenv(
        arguments.name,
        arguments.virtualenvdir,
        python_interpreter=arguments.python,
        user=arguments.virtualenvuser,
        verbose=verbose
    )

    deps = config['pip']['deps']
    use_index = True
    if verbose:
        display_header('Installing python package dependencies')
    try:
        install_python_dependencies(
            virtualenv=virtualenv,
            requirements=arguments.requirement,
            deps=deps,
            upgrade=arguments.upgrade,
            verbose=verbose,
            pip_version=config['pip']['pip_version'],
            use_index=use_index,
            use_local_wheels=use_local_wheels
        )
    except BuildException:
        logger.exception(
            'Package installation in virtualenv failed, removing virtualenv',
        )
        remove_virtualenv(arguments.name, arguments.virtualenvdir)
        raise BuildException('Package installation in virtualenv failed')

    if not arguments.virtualenvuser and not arguments.virtualenvgroup:
        # User didn't specify a user or group
        return

    # Fixing ownership requires both user and group so make sure both
    # are populated if the user didn't pass them.
    if not arguments.virtualenvuser:
        arguments.virtualenvuser = getpass.getuser()

    if not arguments.virtualenvgroup:
        arguments.virtualenvgroup = getpwnam(arguments.virtualenvuser).pw_gid

    if verbose:
        display_header('Fixing file ownership')
    fix_file_ownership(virtualenv, arguments.virtualenvuser, arguments.virtualenvgroup)

    return virtualenv


def deployed_bin_files(venv):
    """
    Gets files that where deployed to the bin directory of the virtualenv.

    Parameters
    ----------
    venv : str
        Path the to virtualenv to do this bin_file links for

    Returns
    -------
    dict
        Key = filename
        Value = sha256 hash
    """
    linked_files = []
    confdir = os.path.join(venv, 'conf')
    before_files_conf = os.path.join(confdir, 'binfiles_predeploy.json')
    after_files_conf = os.path.join(confdir, 'binfiles_postdeploy.json')
    if not os.path.exists(before_files_conf):
        logger.debug('No binfiles_predeploy.json is present')
        return linked_files
    if not os.path.exists(after_files_conf):
        logger.debug('No binfiles_postdeploy.json is present')
        return linked_files

    with open(before_files_conf) as handle:
        before_files = json.load(handle)
    with open(after_files_conf) as handle:
        after_files = json.load(handle)

    before = set(before_files.keys())
    after = set(after_files.keys())

    deployed_files = list(after.difference(before))
    result = {}
    for filename in deployed_files:
        result[filename] = after_files[filename]
    return result


def link_deployed_bin_files(venv, destbin):  # pragma: no cover
    """
    Link the deployed files from the venv bin directory into the destbin
    directory.

    Writes a json list of the created links to conf/created_links.json in
    the virtualenv.

    Parameters
    ----------
    venv: str
        Path to the python virtualenv

    destbin: str
        Destination directory when the links should be

    Returns
    -------
    list
        Full path to files linked
    """
    linked_files = []
    venv_bin = os.path.join(venv, 'bin')
    venv_conf = os.path.join(venv, 'conf')
    linked_files_conf = os.path.join(venv_conf, 'created_links.json')
    for filename in deployed_bin_files(venv):
        source_filename = os.path.expanduser(os.path.join(venv_bin, filename))
        dest_filename = os.path.expanduser(os.path.join(destbin, filename))
        if os.path.exists(dest_filename):
            logger.debug('Not overwriting existing file %r', dest_filename)
            continue

        # os.path.exists() returns False for broken symlinks so we need to check
        # to see if the dest_filename is a symlink.
        if os.path.islink(dest_filename):
            logger.warning('Removing broken symlink %r', dest_filename)
            os.unlink(dest_filename)
        os.symlink(source_filename, dest_filename)
        linked_files.append(dest_filename)
    with open(linked_files_conf, 'w') as handle:
        json.dump(linked_files, handle)


def unlink_deployed_bin_files(venv):  # pragma: no cover
    """
    Undo/delete the symlinks created by link_deployed_bin_files()

    Parameters
    ----------
    venv:str
        Path to the root of the virtualenv
    """
    venv_conf = os.path.join(venv, 'conf')
    linked_files_conf = os.path.join(venv_conf, 'created_links.json')
    if not os.path.exists(linked_files_conf):
        logger.debug('No symlinked files for this virtualenv')
        return
    with open(linked_files_conf, 'r') as handle:
        for filename in json.load(handle):
            if os.path.islink(filename):
                logger.debug('Removing symlink %r', filename)
                os.remove(filename)
                continue
            if os.path.exists(filename):
                logger.debug('File %r is not a symlink, not removing', filename)
                continue
            logger.debug('File %r is missing, cannot remove symlink', filename)
    os.remove(linked_files_conf)
